# -*- coding: utf-8 -*-

from .structures import Codes

status = Codes(
    NO_ALARM = 0,
    READ_ALARM = 1,
    WRITE_ALARM = 2,
    HIHI_ALARM = 3,
    HIGH_ALARM = 4,
    LOLO_ALARM = 5,
    LOW_ALARM = 6,
    STATE_ALARM = 7,
    COS_ALARM = 8,
    COMM_ALARM = 9,
    TIMEOUT_ALARM = 10,
    HWLIMIT_ALARM = 11,
    CALC_ALARM = 12,
    SCAN_ALARM = 13,
    LINK_ALARM = 14,
    SOFT_ALARM = 15,
    BAD_SUB_ALARM = 16,
    UDF_ALARM = 17,
    DISABLE_ALARM = 18,
    SIMM_ALARM = 19,
    READ_ACCESS_ALARM = 20,
    WRITE_ACCESS_ALARM = 21
)

severity = Codes(
    NO_ALARM = 0,
    MINOR = 1,
    MAJOR = 2,
    INVALID = 3,
    EST_REPEAT = 3968,
    REPEAT = 3856,
    DISCONNECTED = 3904,
    ARCHIVE_OFF = 3872,
    ARCHIVE_DISABLED = 3848,
)

interpolation = Codes(
    RAW = 0,
    SPREADSHEET = 1,
    AVERAGED = 2,
    PLOT_BINNING = 3,
    LINEAR = 4,
)

data_type = Codes(
    STRING = 0,
    ENUM = 1,
    INT = 2,
    DOUBLE = 3,
)

xmlrpc = Codes(
    UNSPECIFIED = 0,
    INTERNAL = -500,
    TYPE = -501,
    INDEX = -502,
    PARSE = -503,
    NETWORK = -504,
    TIMEOUT = -505,
    NO_SUCH_METHOD = -506,
    REQUEST_REFUSED = -507,
    INTROSPECTION_DISABLED = -508,
    LIMIT_EXCEEDED = -509,
    INVALID_UTF8 = -510,
)

archiver = Codes(
    SEVER_FAULT = -600,
    NO_INDEX = -601,
    ARGUMENT_ERROR = -602,
    DATA_ERROR = -603,
)



# -*- coding: utf-8 -*-

from collections import namedtuple
from datetime import datetime

from . import codes
from . import utils
from . import exceptions


HAS_NUMPY = False
try:
    import numpy as np
    HAS_NUMPY = True
except ImportError:
    pass


ArchiveProperties = namedtuple('ArchiveProperties',
                               'key start_time end_time')


Limits = namedtuple('Limits', 'low high')


class ChannelData(object):
    '''
    Container for archive data for a single channel.

    Fields:
    channel: The channel name.
    values: A list of channel values.
    times: A list of datetimes corresponding with the retrieved times.
    statuses: A list of status values corresponding with the retrieved
        times.
    severities: A list of severity values corresponding with the
        retrieved times.
    units: The units of the values.
    states: A list of strings describing the states a STRING or ENUM
        can have.
    data_type: The data type of the channel.
    elements: The number of elements per sample for waveform channels.
    display_limits: Values advising how to display the values in a user
        interface.
    warn_limits: Low and high values for which the channel will
        generate a warning.
    alarm_limits: Low and high values for which the channel will
        generate an alarm.
    display_precision: The number of decimal places to show in user
        interfaces.
    archive_key: The archive the data was pulled from.
    '''

    def __init__(self, channel=None, values=None, times=None, statuses=None,
                 severities=None, units=None, states=None, data_type=None,
                 elements=None, display_limits=None, warn_limits=None,
                 alarm_limits=None, display_precision=None, archive_key=None,
                 interpolation=None):

        super(ChannelData, self).__init__()

        self.channel = channel
        self.values = values
        self.times = times
        self.statuses = statuses
        self.severities = severities
        self.units = units
        self.states = states
        self.data_type = data_type
        self.elements = elements
        self.display_limits = display_limits
        self.warn_limits = warn_limits
        self.alarm_limits = alarm_limits
        self.display_precision = display_precision
        self.archive_key = archive_key
        self.interpolation = interpolation
        self._array = None

    @property
    def array(self):
        '''Return the data in a numpy array structure.'''

        if not HAS_NUMPY:
            raise exceptions.NumpyNotInstalled('Numpy not found.')

        # Only compute the array once
        if self._array is None:

            if self.data_type == codes.data_type.STRING:
                value_dtype = np.str
            elif self.data_type == codes.data_type.ENUM:
                value_dtype = np.uint8
            elif self.data_type == codes.data_type.INT:
                value_dtype = np.int
            else:
                value_dtype = np.dtype(float)

            dtypes = [('time', np.dtype('datetime64[us]')),
                      ('value', value_dtype, self.elements),
                      ('status', np.uint8),
                      ('severity', np.uint16)]

            data = zip(self.times, self.values,
                       self.statuses, self.severities)

            self._array = np.array(data, dtype=dtypes)

        return self._array

    def __repr__(self):

        if self.data_type == codes.data_type.DOUBLE:
            fmt = '{0:.6g}'
        else:
            fmt = '{0!r}'

        s = 'ChannelData(\n'
        if self.elements == 1:
            s += utils.pretty_list_repr(self.values, fmt,
                                        prefix='    values=')
        else:
            s += utils.pretty_waveform_repr(self.values, fmt,
                                            prefix='    values=')
        s += ',\n'
        for attr in ['times', 'statuses', 'severities', 'states']:
            value = self.__getattribute__(attr)
            if value is None:
                continue
            prefix = '    {0}='.format(attr)
            s += utils.pretty_list_repr(value, prefix=prefix)
            s += ',\n'
        for attr in ['units', 'data_type', 'elements', 'display_limits',
                     'warn_limits', 'alarm_limits', 'display_precision',
                     'archive_key', 'interpolation']:
            value = self.__getattribute__(attr)
            if value is None:
                continue
            s += '    {0}={1!r},\n'.format(attr, value)
        s = s[:-2]
        s += '\n)'
        return s

    def __str__(self):
        times = ['time'] + [dt.strftime('%Y-%m-%d %H:%M:%S')
                            for dt in self.times]
        statuses = ['status'] + \
                list(map(codes.status.str_value, self.statuses))
        severities = ['severity'] + \
                list(map(codes.severity.str_value, self.severities))
        times_len = len(max(times, key=len))
        statuses_len = len(max(statuses, key=len))
        severities_len = len(max(severities, key=len))
        s = ''
        value_format = '{0:.9g}'
        if self.elements == 1:
            values = ['value'] + list(map(value_format.format, self.values))
        else:
            len_for_values = 79 - times_len - statuses_len - severities_len - 6
            values = ['value']
            max_value_len = utils.max_value_len_in_waveform(self.values,
                                                            value_format)
            for value in self.values:
                formatted_value = utils.pretty_list_repr(
                                    value, value_format,
                                    max_line_len=len_for_values,
                                    min_value_len=max_value_len)
                values += formatted_value.split('\n')

        values_len = len(max(values, key=len))
        spec = ('{0:>' + str(times_len) + '}  '
                '{1:>' + str(values_len) + '}  '
                '{2:>' + str(statuses_len) + '}  '
                '{3:>' + str(severities_len) + '}\n')

        if self.elements == 1:
            for fields in zip(times, values, statuses, severities):
                s += spec.format(*fields)
        else:
            i = 0
            for line in values:
                if i == 0 or '[' in line:
                    s += spec.format(times[i], line, statuses[i], severities[i])
                    i += 1
                else:
                    s += spec.format('', line.ljust(values_len), '', '')
        return s.rstrip()



# -*- coding: utf-8 -*-

import datetime
import calendar
import re
from tzlocal import get_localzone

try:
    StrType = basestring
except NameError: # Python 3
    StrType = str

SECONDS_PER_MINUTE = 60
MINUTES_PER_HOUR = 60
HOURS_PER_DAY  = 24
SECONDS_PER_HOUR = MINUTES_PER_HOUR * SECONDS_PER_MINUTE
SECONDS_PER_DAY = HOURS_PER_DAY * SECONDS_PER_HOUR


class UTC(datetime.tzinfo):
    '''UTC timezone with optional offset.'''

    def __init__(self, offset=0):
        '''
        offset: Timezone offset in hours relative to UTC.
        '''

        if not -24 < offset < 24:
            raise ValueError('offset must be greater than -24 '
                             'and less than 24')

        self.offset = datetime.timedelta(hours=offset)

    def utcoffset(self, dt):
        return self.offset

    def dst(self, dt):
        return datetime.timedelta(0)

    def tzname(self, dt=None):
        total_seconds = SECONDS_PER_DAY * self.offset.days + self.offset.seconds
        if total_seconds == 0:
            return 'UTC'

        sign = '-' if total_seconds < 0 else '+'
        total_seconds = abs(total_seconds)
        components = {
            'sign': sign,
            'hours': int(total_seconds / SECONDS_PER_HOUR),
            'minutes': int((total_seconds % SECONDS_PER_HOUR)/SECONDS_PER_MINUTE),
            'seconds': total_seconds % SECONDS_PER_MINUTE
        }
        if components['seconds']:
            spec = 'UTC{sign}{hours:02d}:{minutes:02d}:{seconds:02d}'
        else:
            spec = 'UTC{sign}{hours:02d}:{minutes:02d}'
        return spec.format(**components)

    def localize(self, dt):
        return dt.replace(tzinfo=self)

    def __str__(self):
        return self.tzname()

    def __repr__(self):
        total_hours = HOURS_PER_DAY * self.offset.days + \
                      float(self.offset.seconds) / SECONDS_PER_HOUR
        return 'UTC({0:+.8g})'.format(total_hours) if total_hours else 'UTC()'


def localize_datetime(dt, tz):
    '''
    Localize a timezone naive datetime. Handles pytz timezones correctly.
    '''

    try:
        return tz.localize(dt)
    except AttributeError:
        return dt.replace(tzinfo=tz)


def datetime_from_isoformat(iso_str):
    '''
    Convert a string in ISO 8601 format into a datetime.
    iso_str: String in ISO 8601 format. Examples:
        '2013-08-19T14:29Z', '2013-08-19 14:29+10:00'
    '''

    match = re.match('^(.*:.*)([\+\-])(\d\d)(?::(\d\d))?$', iso_str)
    if match is not None:
        dt_str, tz_sign, tz_hr, tz_min = match.groups(0)
        offset = int(tz_hr) + int(tz_min)/60.
        if tz_sign == '-':
            offset *= -1
        tz = UTC(offset)
    else:
        if iso_str.endswith('Z'):
            tz = utc
        else:
            tz = local_tz
        dt_str = iso_str.rstrip('Z')
    dt_str = dt_str.replace('T', ' ')

    formats = ['%Y-%m-%d %H:%M:%S.%f',
               '%Y-%m-%d %H:%M:%S',
               '%Y-%m-%d %H:%M',
               '%Y-%m-%d %H',
               '%Y-%m-%d',
               '%Y-%m',
               '%Y']

    dt = None
    for f in formats:
        try:
            dt = datetime.datetime.strptime(dt_str, f)
        except ValueError:
            continue
        else:
            break
    if dt is None:
        raise ValueError(('{0} is not a recognized '
                          'datetime format').format(iso_str))
    dt = tz.localize(dt)
    return dt


def datetime_from_sec_and_nano(seconds, nanoseconds=0, tz=None):
    '''
    Convert seconds and nanoseconds since the Epoch into a datetime with given timezone.
    If tz is not specified, the local timezone will be used.
    Note: Some precision will be lost as datetimes only store microseconds.
    '''
    if tz is None:
        tz = local_tz
    # We create the datetime in two steps to avoid the weird
    # microsecond rounding behaviour in Python 3.
    dt = datetime.datetime.fromtimestamp(seconds, utc)
    if nanoseconds < 999999500:
        dt = dt.replace(microsecond=int(round(1.e-3 * nanoseconds)))
    else:
        dt = datetime.datetime.fromtimestamp(seconds+1, utc)
    return dt.astimezone(tz)


def sec_and_nano_from_datetime(dt):
    '''
    Convert a datetime to seconds and nanoseconds since the Epoch.
    '''
    seconds = int(calendar.timegm(dt.utctimetuple()))
    nanoseconds = int(dt.microsecond * 1e3)
    return seconds, nanoseconds


def overlap_between_intervals(first_range_start, first_range_end,
                              second_range_start, second_range_end):

    latest_start = max(first_range_start, second_range_start)
    earliest_end = min(first_range_end, second_range_end)

    return max(earliest_end - latest_start, datetime.timedelta(0))


def pretty_list_repr(lst, value_format='{0!r}', max_line_len=79, prefix='',
                     min_value_len=0):
    if not lst:
        return repr(lst)
    lst = list(map(value_format.format, lst))
    values = len(lst)
    max_value_len = max(len(max(lst, key=len)), min_value_len)
    prefix_len = len(prefix)
    delim = ', '
    delim_len = len(delim)
    paren_len = 2
    chars_per_line = max_line_len + delim_len - prefix_len - paren_len
    chars_per_value = max_value_len + delim_len
    values_per_line = max(int(chars_per_line / chars_per_value), 1)
    lines = int(values / values_per_line)
    if values % values_per_line != 0:
        lines += 1
    space_spec = '{0:>' + str(max_value_len) + '}'
    start_space = ' ' * (prefix_len + 1)
    s = prefix
    for line in range(lines):
        offset = line * values_per_line
        line_lst = lst[offset:(offset+values_per_line)]
        s += '[' if line == 0 else start_space
        s += delim.join(map(space_spec.format, line_lst))
        s += ']' if line == lines - 1 else ',\n'
    return s


def max_value_len_in_waveform(lst, value_format='{0!r}'):
    max_value_len = 0
    for sub_lst in lst:
        sub_lst = map(value_format.format, sub_lst)
        sub_max_val_len = len(max(sub_lst, key=len))
        max_value_len = max(max_value_len, sub_max_val_len)
    return max_value_len


def pretty_waveform_repr(lst, value_format='{0!r}', max_line_len=79, prefix=''):
    max_value_len = max_value_len_in_waveform(lst, value_format)
    s = ''
    for idx, sub_lst in enumerate(lst):
        p = prefix + '[' if idx == 0 else ' ' * (len(prefix) + 1)
        s += pretty_list_repr(sub_lst, value_format, max_line_len, p, max_value_len)
        s += ']' if idx == len(lst) - 1 else ',\n'
    return s


utc = UTC()
local_tz = get_localzone()



# -*- coding: utf-8 -*-

class ChannelNotFound(LookupError):
    '''Channel not found in archives.'''

class ChannelKeyMismatch(IndexError):
    '''There should be the same number of keys as channels'''

class NumpyNotInstalled(ImportError):
    '''Numpy must be installed for this operation.'''
